#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "mplugin.h"

// Maximum number of models that can be replaced
#define MAX_MODEL_REPLACEMENTS 16

// Number of replacement model data read
int g_iModelsread = 0;

// Structure used to store information from individual
// map config files.
struct modreplaceinfo_s {
   char szOriginal[ 32 ];  // Original model name ( ie. "leet" )
   char szReplace[ 32 ];   // Replacement model name ( ie. "onos" )
   bool bWasloaded;        // True if the model was precached
} gModelinfo[ MAX_MODEL_REPLACEMENTS ];

// Converts forward slashes to backslashes so the paths
// work under win32
void LinuxToWinPath( char* pszPath, int nCount ) {
#ifdef _WIN32
   register int _nCount = nCount;   // I wonder if "register" actually does anything these days

   // Chose to use _nCount so hopefully the compiler would choose to
   // compare to registers rather than a memory location on each iteration.
   for ( register int i = 0; i < _nCount; i++ ) {
      // Convert forward slashes to backslashes so
      // this will work under windows
      if ( pszPath[ i ] == '/' )
         pszPath[ i ] = '\\';
   }
#endif
}

// Writes a string into the server log
void WriteStringToLog( const char* pszFmt, ... ) {
   static char szText[ 1024 ];   // Temp buffer
   va_list vaList;               // Variable args list?

   // Convert the format string into something usable then print it in the server log
   va_start( vaList, pszFmt );
      _vsnprintf( szText, sizeof( szText ), pszFmt, vaList );
   va_end( vaList );

   ALERT( at_logged, "%s", szText );
}

// Removes whitespace characters from the given string
void StripWhitespace( char* pszBuffer, int iBuflen ) {
   char* pszTemp = NULL;   // Pointer to dynamically allocated memory
   int iBufpos = 0;        // Position in the passed buffer
   int iTmppos = 0;        // Position in the temp string

   // Allocate memory for a new string and make sure that allocation
   // succeeded.
   // CHANGE:
   // Using calloc instead of malloc to make sure the string is zeroed out
   // before use.
   if ( ( pszTemp = ( char* ) calloc( 1, iBuflen ) ) != NULL ) {
      // Loop through all characters in the passed buffer
      for ( iBufpos = 0; iBufpos < iBuflen; iBufpos++ ) {
         // If this character is not a whitespace character then copy it
         // into the temporary string
         if ( ! isspace( pszBuffer[ iBufpos ] ) ) {
            pszTemp[ iTmppos ] = pszBuffer[ iBufpos ];
            iTmppos++;
         }
      }

      // Copy the temp string back into the passed buffer and
      // free any memory the temp string used
      strncpy( pszBuffer, pszTemp, iBuflen );
      free( pszTemp );
   }
}

// Reads the model config file for the current map
bool ReadModelConfig( void ) {
   char szConfigpath[ _MAX_PATH ];  // Path to configuration file 
   char szGamedir[ _MAX_PATH ];     // Game directory
   char* pEqualsign = NULL;         // Pointer to an equal sign in szLine
   char szLine[ 1024 ];             // Buffer to receive a line from pConfigfile
   FILE* pConfigfile = NULL;        // File stream for the configuration file
   int iLine = 0;                   // Current line being read from

   // Get the mod directory, format the path to the config and fix it for win32 if necissary
   GET_GAME_DIR( szGamedir );
   _snprintf( szConfigpath, sizeof( szConfigpath ), "%s/modchange_cfgs/%s.txt", szGamedir, STRING( gpGlobals->mapname ) );
   LinuxToWinPath( szConfigpath, ( int ) strlen( szConfigpath ) );

   // Attempt to open the config file for this map
   if ( ( pConfigfile = fopen( szConfigpath, "rt" ) ) != NULL ) {
      // Open was successful, loop until we have reached the end of the file
      while ( ! feof( pConfigfile ) ) {
         szLine[ 0 ] = '\0';  // Make the string look empty at first

         // Read in a line from the config file and remove any
         // whitespace characters.
         fgets( szLine, sizeof( szLine ), pConfigfile );

         // Make sure this line actually has something on it
         if ( strlen( szLine ) > 0 ) {
            StripWhitespace( szLine, strlen( szLine ) );

            // A really big hack to come, just make sure it has an equal character for now
            if ( ( pEqualsign = strstr( szLine, "=" ) ) != NULL ) {
               // Turn that equal sign into a null terminator,
               // now string functions can get the first model name.
               *pEqualsign = '\0'; 
               strncpy( gModelinfo[ g_iModelsread ].szOriginal, szLine, strlen( szLine ) );
               
               // Now, we just read from our equal sign pointer +1 to jump over the null terminator
               // into our second model name :)
               strncpy( gModelinfo[ g_iModelsread ].szReplace, pEqualsign + 1, strlen( pEqualsign + 1 ) );

               // Mark the model as been read and increment the global counter
               gModelinfo[ g_iModelsread ].bWasloaded = true;
               g_iModelsread++;
            }
            else {
               // Warning message since all lines need to have an equal sign
               WriteStringToLog( "[ModelChanger]: Warning: Missing equal sign/parse error on line %d of file %s.\n", iLine, szConfigpath );
            }                  
         }

         // On another line now
         iLine++;
      }

      // Close the file and return success!
      fclose( pConfigfile );
      return true;
   }

   // Bad open, ect...
   return false;
}

// Called when an entity is spawning
int DispatchSpawn( edict_t* pEntity ) {
   char szModelpath[ _MAX_PATH ];   // Full path to model

   // Worldspawn spawns once the map has loaded, now is
   // the perfect time to precache things
   if ( FClassnameIs( pEntity, "worldspawn" ) ) {
      // ALWAYS remember to reset read model data every map
      memset( gModelinfo, 0, sizeof( struct modreplaceinfo_s ) * MAX_MODEL_REPLACEMENTS );
      g_iModelsread = 0;

      // Did we read any config data for this map?
      if ( ReadModelConfig( ) ) {
         // Loop through the data we read in
         for ( int i = 0; i < g_iModelsread; i++ ) {
            // Did we get a name for the replacement model earlier?
            if ( gModelinfo[ i ].bWasloaded ) {
               // Format the path so we can precache the model
               _snprintf( szModelpath, sizeof( szModelpath ), "models/player/%s/%s.mdl", gModelinfo[ i ].szReplace, gModelinfo[ i ].szReplace );

               // Attempt to precache the model, if that fails mark the
               // model as not loaded.
               if ( ! PRECACHE_MODEL( ( char* ) STRING( ALLOC_STRING( szModelpath ) ) ) )
                  gModelinfo[ i ].bWasloaded = false;
            }
         }
      }
   }

   RETURN_META_VALUE( MRES_IGNORED, 1 );
}

// Checks if a file exists
// NOTE: USE ONLY FOR FILES AFTER THE MOD ROOT
int DoesFileExist( const char* pszPath ) {
   char szGamedir[ _MAX_PATH ];  // Mod name
   char szNewpath[ _MAX_PATH ];  // Complete path to file
   FILE* pFile = NULL;           // Temporary file pointer

   // Get the mod directory/name and setup the full path
   GET_GAME_DIR( szGamedir );
   _snprintf( szNewpath, sizeof( szNewpath ), "%s/%s", szGamedir, pszPath );

   // Since fopen() is sensitive to slashes make sure they are
   // correct if using win32.
   LinuxToWinPath( szNewpath, ( int ) strlen( szNewpath ) );

   // If the file opened that means it exists, close the file and
   // return 1.
   if ( ( pFile = fopen( szNewpath, "rb" ) ) != NULL ) {
      fclose( pFile );
      return 1;
   }

   return 0;
}

// Get the path to a replacement model if it exists
int GetNewModelPath( const char* pszModel, char* pszNewmodel, size_t lMaxbuflen ) {
   char szNewpath[ _MAX_PATH ];  // Path to new model
   char szTemp[ _MAX_PATH ];     // Temporary path
   int iPos = 0;                 // Position in string

   // TODO:
   // I really should have commented this when I wrote it since
   // I have no idea why it's here.
   for ( ; iPos < strlen( pszModel ); iPos++ ) {
      if ( pszModel[ iPos ] == '/' ) {
         szTemp[ iPos ] = '\0';
         break;
      }

      szTemp[ iPos ] = pszModel[ iPos ];
      szTemp[ iPos + 1 ] = '\0';
   }

   // Format the full path to the new model
   _snprintf( szNewpath, sizeof( szNewpath ), "%s/%s/%s", szTemp, STRING( gpGlobals->mapname ), pszModel + ( strlen( szTemp ) + 1 ) );

   // If the model exists copy it's path into the given buffer and return 1
   if ( DoesFileExist( szNewpath ) ) {
      strncpy( pszNewmodel, szNewpath, lMaxbuflen );
      return 1;
   }

   return 0;
}

// Checks if there is a replacement for the given sound file.
// If there is the function will return 1, 0 otherwise.
int GetNewSoundPath( const char* pszSound, char* pszNewsound, size_t lMaxbuflen ) {
   char szNewpath[ _MAX_PATH ];  // Path to new sound

   // Format the full sound path
   _snprintf( szNewpath, sizeof( szNewpath ), "sound/%s/%s", STRING( gpGlobals->mapname ), pszSound );

   // If the file exists...
   if ( DoesFileExist( szNewpath ) ) {
      // HACKHACKHACK
      // Skip the sound directory in the path
      strncpy( pszNewsound, szNewpath + 6, lMaxbuflen );
      return 1;
   }
   
   return 0;
}

// Called when the mod wants to precache a model
int PrecacheModel( char* pszModel ) {
   char szNewpath[ _MAX_PATH ];  // Path to replacement model

   // If a replacement model exists for this map, return the index of the new model aswell
   // as precache it.
   if ( GetNewModelPath( pszModel, szNewpath, sizeof( szNewpath ) ) ) {
      RETURN_META_VALUE( MRES_SUPERCEDE, PRECACHE_MODEL( ( char* ) STRING( ALLOC_STRING( szNewpath ) ) ) );
   }

   RETURN_META_VALUE( MRES_IGNORED, 1 );
}
   
// Called when the mod wants to precache a sound
int PrecacheSound( char* pszSound ) {
   char szNewpath[ _MAX_PATH ];  // Path to replacement sound

   // If a replacement sound exists precache it instead
   if ( GetNewSoundPath( pszSound, szNewpath, sizeof( szNewpath ) ) ) {
      RETURN_META_VALUE( MRES_SUPERCEDE, PRECACHE_SOUND( ( char* ) STRING( ALLOC_STRING( szNewpath ) ) ) );
   }

   RETURN_META_VALUE( MRES_IGNORED, 1 );
}
   
// Called when an entity is having it's model set
void SetModel( edict_t* pEntity, const char* pszModel ) {
   char szNewpath[ _MAX_PATH ];  // Path to new model

   // If a replacement model exists, set it to use it instead
   if ( GetNewModelPath( pszModel, szNewpath, sizeof( szNewpath ) ) ) {
      SET_MODEL( pEntity, STRING( MAKE_STRING( szNewpath ) ) );
      RETURN_META( MRES_SUPERCEDE );
   }

   RETURN_META( MRES_IGNORED );
}

// Playing a sound
void EmitSound( edict_t* pEntity, int nChan, const char* pszSample, float flVolume, float flAttenuation, int iFlags, int iPitch ) {
   char szNewpath[ _MAX_PATH ];  // Path to replacement sound

   // If a replacement sound exists, play it instead
   if ( GetNewSoundPath( pszSample, szNewpath, sizeof( szNewpath ) ) ) {
      g_engfuncs.pfnEmitSound( pEntity, nChan, STRING( MAKE_STRING( szNewpath ) ), flVolume, flAttenuation, iFlags, iPitch );
      RETURN_META( MRES_SUPERCEDE );
   }
         
   RETURN_META( MRES_IGNORED );
}

// ???
// Maybe the name is description enough?
void EmitAmbientSound( edict_t* pEntity, float* pflOrigin, const char* pszSample, float flVolume, float flAttenuation, int iFlags, int iPitch ) {
   char szNewpath[ _MAX_PATH ];  // Path to replacement sound

   // If a replacement sound exists, play it instead
   if ( GetNewSoundPath( pszSample, szNewpath, sizeof( szNewpath ) ) ) {
      g_engfuncs.pfnEmitAmbientSound( pEntity, pflOrigin, STRING( MAKE_STRING( szNewpath ) ), flVolume, flAttenuation, iFlags, iPitch );
      RETURN_META( MRES_SUPERCEDE );
   }
      
   RETURN_META( MRES_IGNORED );
}

// Called every time the mod wants to set a value
// on the client.
void SetClientKeyValue( int iPlayerindex, char* pInfobuffer, char* pszKey, char* pszValue ) {
   // Is someone trying to set the model for a player?
   if ( ! stricmp( pszKey, "model" ) ) {
      // Loop through all of the model replacements we read in
      for ( int i = 0; i < g_iModelsread; i++ ) {
         // Make sure they were read correctly
         if ( gModelinfo[ i ].bWasloaded ) {
            // Is there a match for the model being set?
            if ( ! stricmp( gModelinfo[ i ].szOriginal, pszValue ) ) {
               // Use the replacement instead
               SET_CLIENT_KEYVALUE( iPlayerindex, pInfobuffer, pszKey, gModelinfo[ i ].szReplace );
               RETURN_META( MRES_SUPERCEDE );
            }
         }
      }
   }

   // We did nothing!
   RETURN_META( MRES_IGNORED );
}

BEGIN_GIVEFNPTRSTODLL( )
END_GIVEFNPTRSTODLL( )

BEGIN_META_ATTACH( )
END_META_ATTACH( )

BEGIN_META_QUERY( )
END_META_QUERY( )

BEGIN_META_DETACH( )
END_META_DETACH( )

BEGIN_GETENGINEFUNCTIONS( )
   HOOK_ENGINE_FN( pfnPrecacheModel, PrecacheModel );
   HOOK_ENGINE_FN( pfnPrecacheSound, PrecacheSound );
   HOOK_ENGINE_FN( pfnEmitSound, EmitSound );
   HOOK_ENGINE_FN( pfnEmitAmbientSound, EmitAmbientSound );
   HOOK_ENGINE_FN( pfnSetModel, SetModel );
   HOOK_ENGINE_FN( pfnSetClientKeyValue, SetClientKeyValue );
END_GETENGINEFUNCTIONS( )

BEGIN_GETENTITYAPI2( )
   HOOK_DLLAPI_FN( pfnSpawn, DispatchSpawn );
END_GETENTITYAPI2( )
